// @abstract(Generators for RL Core)
// @author(Kornel Kisielewicz <kisiel@fulbrightweb.org>)
// @created(March 6, 2005)
// @lastmod(March 6, 2005)
//
// This unit hold's -cell data;

unit rlgen;
interface
uses vutil,vnode, rllevel;

type CellSet   = set of Byte;

procedure GenerateLevel(nLevel : TLevel; const ltype : string);
procedure GenFindTile(Tile : CellSet);
procedure GenSetLevel(nLevel : TLevel);

const TileMaxX = 25;
      TileMaxY = 25;
      TilesInStore       = 50;
      GenScanInvalid     = 50000;
      MaxFeatureCommands = 5;
      specCellEmpty      = 255;
      specCellExit       = 250;
      MaxDirExits        = 6;    {** max exits in one direction!}
      MaxHotSpots        = 1000;
      MonsterGroupNumber = 30;

type TileExit  = record dir : char; ax : byte; ay : byte end;
type TileMap   = array [1..TileMaxX,1..TileMaxY] of Byte;
type TileExits = array [1..10] of TileExit;

var HotSpots   : array[1..MaxHotSpots] of TileExit;
    HotSpotNum : Word;
    HotSpotMax : Word;
    GenX, GenY : Integer;
    PlayerX,PlayerY : Byte;

type TileData  = class(TVObject)
                   Map           : TileMap;
                   Exits         : TileExits;
                   ExNum         : Byte;
                   SizeX,SizeY   : Byte;
                   typ           : char;
                   Wall          : Byte;
                   AreaID        : string[12];
                   ID            : string[12];
                   Commands      : array[1..MaxFeatureCommands] of AnsiString;
                   EntryX,EntryY : Byte;
                   constructor Create;
                   destructor  Destroy; override;
                   procedure FlipX;
                   procedure FlipY;
                   procedure RandomFlip;
                   procedure FlipExit(var AEx : TileExit);
                   procedure WriteToMap(mx,my : byte);
                   procedure WriteToMapXY(mx,my : byte; cx,cy : byte);
                   function GetExit(var ExitNum : byte; dir : char) : boolean;
                   procedure PickEntry(dir : char);
                   function Scan(mx,my : Integer; GoodTiles : CellSet) : word;
                 end;

type TileStore  = class(TVObject)
                    Data : array[1..TilesInStore] of TileData;
                    Size : word;
                    constructor Init;
                    procedure Load(FName : string);
                    destructor Done;
                  end;

implementation
uses sysutils, vmath, vdebug, vstream, vini, rlglobal, rlnpc, rlmsg, diablout;

const cellFloors = [CellFloor];
      cellWalls  = [CellWall,CellBWall,CellDoorClosed];
      WallCell   : Byte = CellWall;

var Level : TLevel;
    GenOk      : boolean;
    dX, dY     : Integer;
    LimitCount : Word;
    RoomCount  : Word;
    LastRoomX  : Word;
    LastRoomY  : Word;


procedure GenSetLevel(nLevel : TLevel);
begin
  Level := nLevel;
end;

procedure FromFile(FileName : string);
var ISt       : TTextInputStream;
    CharMap   : array[32..128] of Byte;
    ReadChar  : Char;
    ReadStr   : string;
    Count     : byte;
    LineCount : byte;
begin
  ISt := TTextInputStream.Open(FileName);
  while (ISt.ReadString)[1] <> '%' do;
  repeat
    ReadStr := ISt.ReadString;
    if ReadStr[1] = '%' then break;
    ReadChar := ReadStr[1];
    Delete(ReadStr,1,1);
    CharMap[Ord(ReadChar)] := StrToInt(ReadStr);
  until false;
  ISt.Close;
  ISt := TTextInputStream.Open(FileName);
  LineCount := 1;
  repeat
    ReadStr := ISt.ReadString;
    if ReadStr[1] = '%' then break;
    for Count := 1 to Min(length(ReadStr),MapSizeX) do
      Level.Map[Count,LineCount].Cell := CharMap[Ord(ReadStr[Count])];
    Inc(LineCount);
    if LineCount > MapSizeY then break;
  until false;
  ISt.Close;
end;

function SMap(x,y : LongInt) : byte;
begin
  if x < 1 then Exit(255);
  if y < 1 then Exit(255);
  if (x in [1..MapSizeX]) and (y in [1..MapSizeY]) then Exit(Level.Map[x,y].cell);
  Exit(255);
end;

function NewTileExit(d : char; x,y : byte) : TileExit;
begin
  NewTileExit.dir := d;
  NewTileExit.ax  := x;
  NewTileExit.ay  := y;
end;

procedure AddHotSpot(HS : TileExit);
var c : Word;
begin
  for c := 1 to MaxHotSpots do
    if HotSpots[c].dir = 'x' then
      begin
        HotSpots[c] := HS;
        if c > HotSpotMax then HotSpotMax := c;
        Inc(HotSpotNum);
        Exit;
      end;
  CritError('HotSpot Overflow!');
end;

function RandomHotSpot : TileExit;
var c : word;
begin
  if (HotSpotNum = 0) or (HotSpotMax = 0) then Exit;
  repeat
    c := Random(HotSpotMax)+1;
  until HotSpots[c].dir <> 'x';
  if c = HotSpotMax then Dec(HotSpotMax);
  Dec(HotSpotNum);
  RandomHotSpot := HotSpots[c];
  HotSpots[c].dir := 'x';
end;


function GetRanDirection : char;
begin
  if SMap(GenX,GenY+1) in CellFloors then  Exit('n');
  if SMap(GenX,GenY-1) in CellFloors then  Exit('s');
  if SMap(GenX+1,GenY) in CellFloors then  Exit('w');
  if SMap(GenX-1,GenY) in CellFloors then  Exit('e');
  Exit('n');
end;

function GenCorridorGood(a,b:byte):boolean;
begin
  GenCorridorGood := false;
  if (Level.Map[a,b+1].cell in CellFloors) and (Level.Map[a,b-1].cell in CellFloors) then Exit(True);
  if (Level.Map[a+1,b].cell in CellFloors) and (Level.Map[a-1,b].cell in CellFloors) then Exit(True);
end;

constructor TileData.Create;
var cn : byte;
begin
  for cn := 1 to MaxFeatureCommands do Commands[cn] := '';
end;

procedure TileData.PickEntry(dir : char);
type Coord = record x,y : Byte; end;
var dx, dy : ShortInt;
    sx, sy : Byte;
    exlist : array[1..MaxDirExits] of Coord;
    excnt  : byte;
function Good : boolean;
begin
  if sx+dx < 1 then Exit(false);
  if sy+dy < 1 then Exit(false);
  if sx+dx > SizeX then Exit(false);
  if sy+dy > SizeY then Exit(false);
  if (Map[sx+dx,sy+dy] <> SpecCellEmpty) and
     (not (Map[sx+dx,sy+dy] in CellWalls)) then Exit(true);
  Exit(false);
end;
begin
  dx := 0; dy := 0;

  case dir of
    's','S' : dy := +1;
    'n','N' : dy := -1;
    'e','E' : dx := +1;
    'w','W' : dx := -1;
  end;
  excnt := 0;
  for sx := 1 to SizeX do
    for sy := 1 to SizeY do
      if Map[sx,sy] = SpecCellExit then
        if Good then
          begin
            Inc(excnt);
            if excnt > MaxDirExits then CritError('TileData.PickEntry -- overflow of exits in one direction (MaxDirExits)');
            exlist[excnt].x := sx;
            exlist[excnt].y := sy;
          end;
  if excnt < 1 then CritError('TileData.PickEntry -- Tile "'+ID+'": no exit in "'+dir+'" direction.');
  excnt := Random(excnt)+1;
  EntryX := exlist[excnt].x;
  EntryY := exlist[excnt].y;
end;

destructor  TileData.Destroy;
begin
  inherited Destroy;
end;

constructor TileStore.Init;
var cnt : word;
begin
  for cnt := 1 to TilesInStore do Data[cnt] := nil;
  for cnt := 1 to MaxHotSpots  do HotSpots[cnt].dir := 'x';
  HotSpotNum := 0;
  HotSpotMax := 0;
  Size := 0;
end;

function CVal(ch : char): byte;
begin
  cval := 0;
  if ch in ['1'..'9'] then Exit(StrToInt(ch));
  case ch of
    'a' : cval := 10; 'b' : cval := 11; 'c' : cval := 12;  'd' : cval := 13;
    'e' : cval := 14; 'f' : cval := 15; 'g' : cval := 16;  'h' : cval := 17;
    'i' : cval := 18; 'j' : cval := 19;
    'A' : cval := 20; 'B' : cval := 21; 'C' : cval := 22;  'D' : cval := 23;
    'E' : cval := 24; 'F' : cval := 25; 'G' : cval := 26;  'H' : cval := 27;
    'I' : cval := 28; 'J' : cval := 29;
  end;
end;

function Param(s : string; nr : byte):string;
begin
  Param := Parameter(s,nr,':');
end;

function nParam(s : string; nr : byte):word;
begin
  nParam := StrToInt(Parameter(s,nr,':'));
end;

function cParam(s : string; nr : byte):char;
begin
  cParam := Parameter(s,nr,':')[1];
end;


procedure TileStore.Load(FName : string);
var Count : word;
    cn,cc : byte;
    VF    : TTextInputStream;
    VS    : string;
begin
  VF := TTextInputStream.Open(Fname);
  repeat
    VS := VF.ReadString;
    if (VS[1] = '<') then if (Parameter(VS,1,'<>') <> '') then
      if (VS[2] <> '/') then
        if Parameter(VS,1,'<>') <> 'end' then
       begin
         Inc(Size);
         if Size > TilesInStore then
            CritError('Tiles overflow, while loading "'+FName+'"!');
         Data[Size] := TileData.Create;
         with Data[Size] do
         begin
           ID     := Parameter(VS,1,'<>');
           VS     := VF.ReadString;
           if Param(VS,1) <> '#gen' then CritError('Bad feature data, feature "'+ID+'" in '+FName+'.');
           SizeX  := nParam(VS,2);
           SizeY  := nParam(VS,3);
           Typ    := cParam(VS,4);
           AreaID :=  Param(VS,5);
           Wall   := nParam(VS,6);
           for Count := 1 to SizeY do
           begin
             VS := VF.ReadString;
             if VS[1] = '$' then CritError('Error in MapSizes, in TileFile "'+FName+'"!');
             for cn := 1 to SizeX do
               if VS[cn] = '.' then Map[cn,count] := specCellEmpty else
                 if VS[cn] = 'x' then Map[cn,count] := specCellExit else
                   Map[cn,count] := CVal(VS[cn]);
           end;
           ExNum := 0;
           repeat
             VS := VF.ReadString;
             if VS[1] = '!' then
             begin
               Inc(ExNum);
               Exits[ExNum].Dir := cParam(VS,2);
               Exits[ExNum].ax  := nParam(VS,3);
               Exits[ExNum].ay  := nParam(VS,4);
             end;
           until VS[1] <> '!';
           cc := 0;
           repeat
             if  VS[1] <> '<' then
             begin Inc(cc);
               if cc > MaxFeatureCommands then CritError('Feature Commands Overflow -- feature "'+ID+'" in '+FName+'.');
               Commands[cc] := VS;
               VS := VF.ReadString;
             end;
           until VS[1] = '<';
         end;
       end;
  until VS = '<end>';
  VF.Close;
end;

destructor TileStore.Done;
var cnt : word;
begin
  for cnt := 1 to TilesInStore do
    if Data[cnt] <> nil then FreeAndNil(Data[cnt]);
  Size := 0;
end;

procedure TileData.FlipX;
var Temp : TileData;
    TX,TY : Byte;
begin
  Temp := Self;
  for TX := 1 to SizeX do
    for TY := 1 to SizeY do
      Temp.Map[TX,TY] := Map[SizeX-TX+1,TY];
  for TY := 1 to 10 do if Temp.Exits[TY].Dir in ['w','e'] then FlipExit(Temp.Exits[TY]);
  Self := Temp;
end;

procedure TileData.FlipY;
var Temp : TileData;
    TX,TY : Byte;
begin
  Temp := Self;
  for TX := 1 to SizeX do
    for TY := 1 to SizeY do
      Temp.Map[TX,TY] := Map[TX,SizeY-TY+1];
  for TY := 1 to 10 do if Temp.Exits[TY].Dir in ['n','s'] then FlipExit(Temp.Exits[TY]);
  Self := Temp;
end;

procedure TileData.RandomFlip;
begin
  case Random(4) of
    1 : FlipX;
    2 : FlipY;
    3 : begin FlipX; FlipY; end;
  end;
end;

procedure TileData.WriteToMap(mx,my : byte);
var a,b : byte;
begin
  for a := 1 to SizeX do
    for b := 1 to SizeY do
      if Map[a,b] <> specCellEmpty then
        if SMap(mx+a-1,my+b-1) <> 255 then
          if Map[a,b] = specCellExit then
          begin
            Level.Map[mx+a-1,my+b-1].cell := Wall;
            AddHotSpot(NewTileExit('z',mx+a-1,my+b-1));
          end
             else
               Level.Map[mx+a-1,my+b-1].Cell := Map[a,b];
  Inc(RoomCount);
end;

procedure TileData.WriteToMapXY(mx,my : byte; cx,cy : Byte);
var a,b : byte;
begin
  for a := 1 to SizeX do
    for b := 1 to SizeY do
      if Map[a,b] <> specCellEmpty then
        if SMap(mx+a-cx,my+b-cy) <> 255 then
          if Map[a,b] = specCellExit then
          begin
            Level.Map[mx+a-cx,my+b-cy].cell := Wall;
            AddHotSpot(NewTileExit('z',mx+a-cx,my+b-cy));
          end
             else
               Level.Map[mx+a-cx,my+b-cy].cell := Map[a,b];
  Inc(RoomCount);
end;


function TileData.GetExit(var ExitNum : byte; dir : char) : boolean;
var cnt : byte;
begin
  cnt := 0;
  GetExit := False;
  repeat Inc(cnt); until (Exits[cnt].dir = dir) or (cnt = 10);
  if Exits[cnt].dir = dir then
  begin
    exitNum := cnt;
    GetExit := True;
  end;
end;


function DirRev(dir : char):char;
begin
  case dir of
    'n' : Exit('s');
    's' : Exit('n');
    'e' : Exit('w');
    'w' : Exit('e');
  end;
end;

procedure TileData.FlipExit(var AEx : TileExit);
begin
  AEx.Dir := DirRev(AEx.dir);
end;


procedure GenRandomTile(Tile : CellSet);
const Limit = 20000;
begin
  LimitCount := 0;
  repeat
    Inc(LimitCount);
    GenX := Random(MapSizeX-2)+1;
    GenY := Random(MapSizeY-2)+1;
  until (Level.Map[GenX,GenY].cell in Tile) or (LimitCount > Limit);
  if LimitCount > Limit then Warning('GenRandomTile -- Failed!');
end;

procedure GenFindTile(Tile : CellSet);
var c,d : byte;
begin
  for c := 1 to MapSizeX do
   for d := 1 to MapSizeY do
     if Level.Map[c,d].cell in Tile then
       begin GenX := c; GenY := d; Exit; end;
  Warning('GenFindTile -- Failed!');
end;


function GenScanSpace(x1,y1,x2,y2 :Integer):Word;
{zwraca ilo innych pl}
var Other : word;
    a,b   : Word;
begin
  if SMap(x1,y1) = 255 then Exit(GenScanInvalid);
  if SMap(x2,y2) = 255 then Exit(GenScanInvalid);
  if not ((x1 in [2..MapSizeX]) and (x2 in [2..MapSizeX]) and
     (y1 in [2..MapSizeY]) and (y2 in [2..MapSizeY])) then Exit(GenScanInvalid);

  Other := 0;
  for a := x1 to x2 do
   for b := y1 to y2 do
     if not (SMap(a,b) in CellWalls) then Inc(Other);
  GenScanSpace := Other;
end;

function TileData.Scan(mx,my : Integer; GoodTiles : CellSet):Word;
{zwraca ilo innych pl}
var Other : word;
    a,b   : Word;
begin
  if SMap(mx,my) = 255 then Exit(GenScanInvalid);
  if SMap(mx+SizeX-1,my+SizeY-1) = 255 then Exit(GenScanInvalid);
  if not ((mx in [2..MapSizeX]) and (mx+SizeX-1 in [2..MapSizeX]) and
     (my in [2..MapSizeY]) and (my+SizeY-1 in [2..MapSizeY])) then Exit(GenScanInvalid);

  Other := 0;
  for a := mx to mx+SizeX-1 do
    for b := my to my+SizeY-1 do
      if Map[a-mx+1,b-my+1] <> specCellEmpty then
        if not (SMap(a,b) in GoodTiles) then Inc(Other);
  Scan := Other;
end;


procedure GenCreateRoom(x1,y1,x2,y2,fill : byte);
var a,b : Word;
begin
  LastRoomX := x1;
  LastRoomY := y1;
  for a := x1 to x2 do
   for b := y1 to y2 do
     Level.Map[a,b].cell := fill;
end;

procedure GenCreateHall(x1,y1,x2,y2,fill : byte);
var a,b : Word;
begin
  for a := x1 to x2 do
   for b := y1 to y2 do
    Level.Map[a,b].cell := fill;
end;

procedure GenCreateRoughHall(x1,y1,x2,y2,Wide, fill : byte; axis : char; ChangeRate : byte);
var a,b   : Word;
    Shift : ShortInt;
begin
  Shift := 0;
  for a := x1 to x2 do
   for b := y1 to y2 do
    begin
      case axis of
        'y' : Level.Map[a+Shift,b].cell := fill;
        'x' : Level.Map[a,b+Shift].cell := fill;
      end;
      if (a <> x1) or (b <> y1) then
      begin
        case Random(ChangeRate+1) of
          1 : if Shift <  (Wide-1) then Inc(Shift);
          2 : if Shift > -(Wide-1) then Dec(Shift);
        end;
        case axis of
          'y' : Level.Map[a+Shift,b].cell := fill;
          'x' : Level.Map[a,b+Shift].cell := fill;
        end;
      end;
    end;
  case axis of
    'y' : begin GenX := a+Shift; GenY := b; end;
    'x' : begin GenX := a; GenY := b+Shift; end;
  end;
end;


function GenAround(a,b : byte; fill : CellSet):byte;
var num : byte;
begin
  num := 0;
  if SMap(a+1,b) in fill then Inc(num);
  if SMap(a-1,b) in fill then Inc(num);
  if SMap(a,b+1) in fill then Inc(num);
  if SMap(a,b-1) in fill then Inc(num);
  if SMap(a+1,b+1) in fill then Inc(num);
  if SMap(a-1,b+1) in fill then Inc(num);
  if SMap(a+1,b-1) in fill then Inc(num);
  if SMap(a-1,b-1) in fill then Inc(num);
  GenAround := num;
end;

function GenCrossAround(a,b : byte; fill : CellSet):byte;
var num : byte;
begin
  num := 0;
  if SMap(a,b+1) in fill then begin Inc(num); dy := 1; end;
  if SMap(a,b-1) in fill then begin Inc(num); dy := -1; end;
  if SMap(a+1,b) in fill then begin Inc(num); dx := 1; end;
  if SMap(a-1,b) in fill then begin Inc(num); dx := -1; end;
  GenCrossAround := num;
end;

function GenArcAround(a,b:byte; da, db : shortint; fill : CellSet):byte;
var num : byte;
begin
  num := 0;
  if da <> 0 then
  begin
    if SMap(a+da,b-1) in fill then Inc(num);
    if SMap(a+da,b  ) in fill then Inc(num);
    if SMap(a+da,b+1) in fill then Inc(num);
  end;
  if db <> 0 then
  begin
    if SMap(a+1,b+db) in fill then Inc(num);
    if SMap(a  ,b+db) in fill then Inc(num);
    if SMap(a-1,b+db) in fill then Inc(num);
  end;
  GenArcAround := num;
end;

procedure GenFindBigEmptySpace;
const CriticalSize = 6000;
var Critical : word;
begin
  Critical := 0;
  repeat
    Inc(Critical);
    GenX := Random(MapSizeX-2)+1;
    GenY := Random(MapSizeY-2)+1;
  until ((SMap(GenX,GenY) in CellFloors) and (GenAround(GenX,GenY,CellFloors) = 8))
        or (Critical > CriticalSize);
  if Critical > CriticalSize then
  begin
    GenFindTile(CellFloors);
    Warning('FindBigEmptySpace -- Failed!');
  end;
end;

procedure GenFindTotalyEmptySpace;
const CriticalSize = 6000;
var Critical : word;
begin
  Critical := 0;
  repeat
    repeat
      Inc(Critical);
      GenX := Random(MapSizeX-2)+1;
      GenY := Random(MapSizeY-2)+1;
    until ((SMap(GenX,GenY) in CellFloors))
          or (Critical > CriticalSize);
    if Critical > CriticalSize then
    begin
      GenFindTile(CellFloors);
      Warning('FindBigEmptySpace -- Failed!');
    end;
   until (Level.Map[GenX,GenY].NPC = nil) and (Level.Map[GenX,GenY].Item = nil)
end;

function GenFindNearSpace(Range : Byte) : boolean;
const CriticalSize = 100;
var Critical : word;
    GX,GY : Integer;
begin
  Critical := 0;
  repeat
    repeat
      Inc(Critical);
      GX := Random(2*Range+1)-Range+GenX;
      GY := Random(2*Range+1)-Range+GenY;
    until (SMap(GX,GY) in CellFloors) or (Critical > CriticalSize);
    if Critical > CriticalSize then Exit(false);
  until (Level.Map[GX,GY].NPC = nil) and (Level.Map[GX,GY].Item = nil);
  GenX := GX;
  GenY := GY;
  Exit(true);
end;


procedure GenPlaceDownStaircase(n : byte);
begin
  GenFindBigEmptySpace;
  Level.Map[GenX,GenY].cell := n;
end;

procedure GenPlaceUpStaircase(n : byte);
begin
  GenFindBigEmptySpace;
  Level.Map[GenX,GenY].cell := n;
end;

procedure GenSprinkle(tile : byte; num : word);
var cnt : word;
begin
  for cnt := 1 to num do
    Level.Map[Random(MapSizeX)+1,Random(MapSizeY)+1].cell := tile;
end;

procedure GenFill(tile : byte);
var x,y : Word;
begin
  for x := 1 to MapSizeX do
    for y := 1 to MapSizeY do
      Level.Map[x,y].cell := tile;
end;

procedure GenRestoreWalls(Tile : byte);
var mx,my : Byte;
begin
  for mx := 1 to MapSizeX do
    Level.Map[mx,1].cell := Tile;
  for mx := 1 to MapSizeX do
    Level.Map[mx,MapSizeY].cell := Tile;
  for my := 1 to MapSizeY do
    Level.Map[1,my].cell := Tile;
  for my := 1 to MapSizeY do
    Level.Map[MapSizeX,my].cell := Tile;
end;

procedure GenRestoreWalls;
begin
  GenRestoreWalls(WallCell);
end;

{*********}


procedure FilterCircles(hits, radius, fill : byte);
var mx,my,k : byte;
    zx,zy   : ShortInt;
begin
  for k := 1 to hits do
  begin
    mx := Random(MapSizeX-1)+1;
    my := Random(MapSizeY-1)+1;
    for zx := -Radius to Radius do
      for zy := -Radius to Radius do
        if Distance(mx,my,mx+zx,my+zy) <= Radius then
          if SMap(mx+zx,my+zy) <> 255 then Level.Map[mx+zx,my+zy].cell := fill;
  end;
end;

function GCmdName(CmdStr : string) : string;
var ComName  : string[80];
    Position : byte;
begin
  ComName := '';
  Position := 1;
  repeat
    ComName := ComName + CmdStr[Position];
    Inc(Position);
  until (Position > Length(CmdStr)) or (CmdStr[Position] = '(') or (CmdStr[Position] = '}');
  GCmdName := LowerCase(ComName);
end;


procedure ClearDeadEnds;
var GenX,GenY : byte;
begin
  for GenX := 2 to MapSizeX-1 do
    for GenY := 2 to MapSizeY-1 do
      if (GenCrossAround(GenX,GenY,CellWalls) = 3) then
        Level.Map[GenX,GenY].cell := WallCell;
end;

{************* CAVES ***************}

procedure GenElipse;
{to be developed}
var tx,ty : byte;
    cx,cy : byte;
    sq_a,
    sq_b  : word;
begin
  GenFill(WallCell);
  cx   := MapSizeX div 2;
  cy   := MapSizeY div 2;
  sq_a := Sqr(cx-3);
  sq_b := Sqr(cy-2);

  for tx := 0 to MapSizeX-1 do
    for ty := 0 to MapSizeY-1 do
      if ((Sqr(tx-cx)/sq_a)+(Sqr(ty-cy)/sq_b)) <= 1 then
         Level.Map[tx+1,ty+1].Cell := CellFloor;
end;

procedure GenLargeCave;
var tx,ty : byte;
    cx,cy : byte;
    sq_a,
    sq_b  : word;
    cnt   : word;
begin
  GenFill(WallCell);
  cx   := MapSizeX div 2;
  cy   := MapSizeY div 2;
  sq_a := Sqr(cx-12);
  sq_b := Sqr(cy-8);

  for tx := 0 to MapSizeX-1 do
    for ty := 0 to MapSizeY-1 do
      if ((Sqr(tx-cx)/sq_a)+(Sqr(ty-cy)/sq_b)) <= 1 then
         Level.Map[tx+1,ty+1].cell := CellFloor;

  sq_a := Sqr(cx-20);
  sq_b := Sqr(cy-10);
  cnt := 1;
  repeat
    repeat
      tx := Random(MapSizeX-2)+2;
      ty := Random(MapSizeY-2)+2;
    until ((Sqr(tx-cx)/sq_a)+(Sqr(ty-cy)/sq_b)) >= 1;
    if GenCrossAround(tx,ty,CellFloors) > 0 then Level.Map[tx,ty].cell := CellFloor;
    Inc(cnt);
  until cnt > 10000;
end;

{CIIIIIIITTTTTTTTYYYYYYYYY***************************}
(*
procedure GenCity_AddSquare;
var x1,x2,y1,y2 : byte;
begin
  x1 := (((Random(MapSizeX)+1) div 12)+2)*4;
  y1 := (((Random(MapSizeY)+1) div 12)+1)*4;
  x2 := (((Random(MapSizeX)+1) div 12)+((MapSizeX div 12)*2)-2)*4;
  y2 := (((Random(MapSizeY)+1) div 12)+((MapSizeY div 12)*2)-1)*4;
  GenCreateRoom(x1,y1,x2,y2,CellWall);
  GenCreateRoom(x1+1,y1+1,x2-1,y2-1,CellFloor);
end;

procedure GenCity_DrawRoad(x1,y1,x2,y2 : byte; doortile : byte);
var tx,ty : byte;
    tw    : ShortInt;
begin
  for tx := x1 to x2 do
    for ty := y1 to y2 do
      begin
        if SMap(tx,ty) = 255 then Continue;
        if Level.Map[tx,ty].cell = CellWall  then Level.Map[tx,ty].cell := doorTile;
        if Level.Map[tx,ty].cell in [11,18]  then Level.Map[tx,ty].cell := 12;
        if Level.Map[tx,ty].cell = CellFloor then Level.Map[tx,ty].cell := 12;
      end;
end;

function LineYScan(LineY : byte; cntCells : CellSet) : byte;
var tx  : byte;
    cnt : byte;
begin
  cnt := 0;
  for tx := 1 to LSizeX do
    if Level.Map.c[tx,LineY] in cntCells then Inc(cnt);
  LineYScan := cnt;
end;

function LineXScan(LineX : byte; cntCells : CellSet) : byte;
var ty  : byte;
    cnt : byte;
begin
  cnt := 0;
  for ty := 1 to LSizeY do
    if Level.Map.c[LineX,ty] in cntCells then Inc(cnt);
  LineXScan := cnt;
end;

procedure GenCityWalls;
var ax,ay : byte;
begin
  GenFill(11);
  GenSprinkle(18,200);
  for ax := 1 to 3 do GenCity_AddSquare;
  for ax := 2 to LSizeX-1 do
    for ay := 2 to LSizeY-1 do
      if Level.Map.c[ax,ay] = CellWall then
        if GenCrossAround(ax,ay,CellFloors) >= 2 then
          if GenAround(ax,ay,[11,18]) = 0 then
             Level.Map.c[ax,ay] := CellFloor;


  {1. Doda drogi wedug ktrych uoone jest miasto.              }
  {2. Zaznaczy drogi = przy przeciciu z murami doda bram ;-)   }
  {TEMP}
  repeat
    ay := Random(LSizeY-20)+10;
  until (LineYScan(ay, [CellWall]) = 2) and (LineYScan(ay+1,[CellWall]) = 2);
  GenCity_DrawRoad(1,ay,LSizeX,ay+1,29);
  if k(1,2) = 1 then
  begin
    repeat
      ax := Random(LSizeX-20)+10;
    until (LineXScan(ax, [CellWall]) = 2) and (LineXScan(ax+1,[CellWall]) = 2);
    GenCity_DrawRoad(ax,1,ax+1,LSizeY,31);
  end;
  {/TEMP}
  {3. Doda wiee na rogach!                                       }
  {TEMP}
  for ax := 2 to LSizeX-1 do
    for ay := 2 to LSizeY-1 do
      if Level.Map.c[ax,ay] = CellWall then
        if GenAround(ax,ay,CellFloors) = 1 then
          if GenAround(ax,ay,CellWalls) = 2 then
             GenCreateRoom(ax-1,ay-1,ax+1,ay+1,CellWall);
  {/TEMP}

  {4. teraz losowym algorytmem doda budynki (scan musi by czysty)}

  GenSprinkleTiles('towntile.dat',[CellFloor],1000);

  Ay := k(1,5)+5;
  for AX := 1 to AY do
  begin
    GenRandomTile(CellFloors+[12]);
  end;
end;

*)

procedure GenSprinkleTiles(TileSetName : string; GoodCells : CellSet; MaxTry : word);
var Tiles : TileStore;
    cnt   : word;
    ran   : word;
    gx,gy : byte;
    sc    : word;
begin
  Tiles := TileStore.Init;
  Tiles.Load(TileSetName);
  for cnt := 1 to MaxTry do
  begin
    ran := Random(Tiles.Size)+1;
    gx  := Random(MapSizeX-2)+1;
    gy  := Random(MapSizeY-2)+1;
    Tiles.Data[ran].RandomFlip;
    sc  := Tiles.Data[ran].Scan(gx,gy,GoodCells);
    if sc = 0 then Tiles.Data[ran].WriteToMap(gx,gy);
  end;
  Tiles.Done;
end;

function RandomDoor : Byte;
begin
  case Random(10) of
    0..2 : Exit(CellDoorClosed);
    3..9 : Exit(CellDoorClosed2);
  end;
end;

procedure GenRoomLevel(RoomFile : string; DoorChance : byte; Spec : byte);
var FCnt    : word;
    FLimit  : word;
    ASS     : TileStore;
    FS      : TileStore;
    RoomCnt : word;
    sca     : word;
    procedure PickHotSpot;
    var TE : TileExit;
    begin
      TE := RandomHotSpot;
      GenX := TE.ax;
      GenY := TE.ay;
    end;
begin
  FS := TileStore.Init;
  FS.Load(RoomFile);
  GenFill(WallCell);
  FCnt := 0;
  FLimit := 300;

  GenX := Random(MapSizeX-30)+15;
  GenY := Random(MapSizeY-14)+7;
  dx := Random(2)-1;
  dy := Random(2)-1;
  RoomCnt := 0;
  for FCnt := 1 to 20 do
    AddHotSpot(NewTileExit('z',GenX,GenY));
  FCnt := 0;

  repeat
    Inc(FCnt);
    with FS.Data[Random(FS.Size)+1] do
    begin
      GenOK := True;
      PickHotSpot;
      RandomFlip;
      PickEntry(GetRanDirection);
      sca := Scan(GenX-EntryX+1,GenY-EntryY+1,CellWalls);
      if sca = GenScanInvalid then Continue;
      if sca > 1 then GenOk := False;
      if not GenOk then Continue;
      WriteToMapXY(GenX,GenY,EntryX,EntryY);
      Level.Map[GenX,GenY].cell := RandomDoor;
    end;
  until FCnt >= FLimit;

  for FCnt := 1 to 10 do ClearDeadEnds;
  FCnt := 0;
  for GenX := 1 to MapSizeX do
  begin
    Inc(FCnt);
    for GenY := 1 to MapSizeY do
     if Level.Map[GenX,GenY].cell in CellFloors then
     if (GenCrossAround(GenX,GenY,CellFloors) > 0) and
       (GenCrossAround(GenX,GenY,CellFloors) = 2) and
         (GenCrossAround(GenX,GenY,CellWalls) = 2) then
       if GenCorridorGood(GenX,GenY) then
         if (GenAround(GenX,GenY,CellFloors) = 4) then
            if Random(100) < DoorChance then
               Level.Map[GenX,GenY].cell := RandomDoor;
    if FCnt mod 10 = 0 then
  end;
  GenRestoreWalls(WallCell);
  if WallCell <> CellWall then
  for GenX := 1 to MapSizeX do
    for GenY := 1 to MapSizeY do
      if Level.Map[GenX,GenY].cell = CellWall then
         Level.Map[GenX,GenY].cell := WallCell;



  if Spec <> 1 then GenPlaceDownStaircase(CellStairDown);
  GenPlaceUpStaircase(CellStairUp);
  RoomCnt := 0;
  FS.Done;
end;

function MonsterAmount(const Monster : string) : Word;
begin
  Database.SetSection(Monster);
  if not Database.Defined('amountmin') then Exit(1);
  if not Database.Defined('amountmax') then Exit(Database.GetNumber('amountmin'));
  Exit(RandRange(Database.GetNumber('amountmin'),Database.GetNumber('amountmax')));
end;

procedure PlaceItems;
var c1,c2 : Word;
begin
  for c1 := 1 to 10 do
  begin
    GenFindBigEmptySpace;
    Level.Map[GenX,GenY].Cell := 20;
    //Level.AddItem('pofhealing',GenX,GenY);
  end;
  for c1 := 1 to 8 do
  begin
    GenFindBigEmptySpace;
    case Random(10) of
      0..5 : Level.Map[GenX,GenY].Cell := 21;
      6..8 : Level.Map[GenX,GenY].Cell := 22;
    end;
  end;
  for c1 := 1 to 8 do
  begin
    GenFindBigEmptySpace;
    for c2 := 1 to Random(5)+1 do
      if GenFindNearSpace(2) then
      begin
        Level.Map[GenX,GenY].Cell := 28; // barrel
      end;
  end
end;

const MonsterMaxSize = 4000-386{GOLEM};
type TMonsterList = array[0..25] of string[30];
var MList : TMonsterList;
    MAmount : Word;

{function RandomMonster : string;
begin
  Database.SetSection('groupml'+NumToStr(Level.Level));
  if Database.Selected.Keys = 0 then CritError('INI File : Group "groupml'+NumToStr(Level.Level)+'" empty!');
  Exit(Database.Selected.RandomItem.ReturnString);
end;}

function RandomMonster : string;
begin
  Exit(MList[Random(MAmount)]);
end;

var MonsterCount : Word;
    MonsterAll   : TMonsterList;

procedure MonGenQuery(const nkey,nvalue : string);
begin
  MonsterAll[MonsterCount] := nvalue;
  Inc(MonsterCount);
end;

procedure GenerateMonsterList;
var Count    : Word;
    Size     : Word;
    Rand     : Word;
    Limit    : Word;
    Current  : Word;
    ListSize : Word;
begin
  Database.SetSection('groupml'+IntToStr(Level.Level));
  MonsterCount := 0;
  Database.QuerySection(@MonGenQuery);
  if MonsterCount = 0 then CritError('INI File : Group "groupml'+IntToStr(Level.Level)+'" empty!');

  Count := 0;
  MAmount := 0;
  Size := MonsterCount;
  Limit := MonsterMaxSize;
  repeat
    repeat Rand := Random(Size); until MonsterAll[Rand] <> '';
    Database.SetSection(MonsterAll[Rand]);
    Inc(Count);
    Current := Database.GetNumber('size');
    if Current <= Limit then
    begin
      Dec(Limit,Current);
      MList[MAmount] := MonsterAll[rand];
      Inc(MAmount);
    end;
    MonsterAll[rand] := '';
  until Count = Size;
  if MAmount = 0 then CritError('No monsters generated for level#'+IntToStr(Level.Level)+'!');
end;

procedure PlaceMonsters;
var c1,c2,a : Word;
    m : string;
    NPC : TNPC;
begin
  GenerateMonsterList;
  for c1 := 1 to MonsterGroupNumber do
  begin
    m := RandomMonster;
    a := MonsterAmount(m);
    GenFindTotalyEmptySpace;
    Level.AddMonster(m,GenX,GenY);
    Log('Added '+m);
    if a = 1 then continue;
    for c2 := 2 to a do
    begin
      if GenFindNearSpace(4) then
      begin
        Level.AddMonster(m,GenX,GenY);
      end;
    end;
  end;
end;

procedure GenerateLevel(nLevel : TLevel; const ltype : string);
begin
  Level := nLevel;
  if ltype = 'file' then
  begin
    FromFile(Database.GetString('filename'));
    PlayerX := 82;
    PlayerY := 52;
    Level.Map[61,48].NPC := TNPC.Create('ogden');
    Level.Map[70,49].NPC := TNPC.Create('griswold');
    Level.Map[84,17].NPC := TNPC.Create('adria');
    Level.Map[78,62].NPC := TNPC.Create('farnham');
    Level.Map[61,60].NPC := TNPC.Create('pepin');
    Level.Map[66,54].NPC := TNPC.Create('cain');
    Level.Map[48,50].NPC := TNPC.Create('gillian');
    Level.Map[28,28].NPC := TNPC.Create('dying');
    DiabloRefreshShops;
    //PlayerX := 30;
    //PlayerY := 23;
  end;
  if ltype = 'room' then
  begin
    WallCell := CellWall;
    GenRoomLevel('BASE.DAT',30,0);
    PlayerX := GenX; PlayerY := GenY;
    PlaceItems;
    PlaceMonsters;
    if Level.Level = 1 then if Random(2) = 1 then Msg.Add('The smell of death surrounds you...')
                                             else Msg.Add('The sanctity of this place has been fowled...');
  end;
  if ltype = 'butcher' then
  begin
    WallCell := CellBWall;
    GenRoomLevel('BASE.DAT',30,1);
    PlayerX := GenX; PlayerY := GenY;
    PlaceItems;
    PlaceMonsters;

    GenFindTotalyEmptySpace;
    Level.AddMonster('butcher',GenX,GenY);
    Msg.Add('The smell of blood everywhere!');
    Msg.Add('The Butcher is loose...');
  end;
end;

end.

