// @abstract(Game Level for RL Core)
// @author(Kornel Kisielewicz <kisiel@fulbrightweb.org>)
// @created(January 17, 2005)
// @lastmod(January 17, 2005)
//
// This unit hold's level class : TLevel.

unit rllevel;
interface
uses vutil, rlgobj, rlglobal, rlnpc, rlthing, rlitem;

// Defines a single map cell. The pointers herin are only informational.
// Disposing of the NPCs and Items is done by disposing TLevel's children.
// Control of the cell pointers is done by the objects themselves (see
// TThing.Displace.
type TCell  = record
       Cell   : Byte;
       Light  : Byte;
       NPC    : TNPC;
       Item   : TItem;
     end;

// The TLevel class. Holds an array of TCells (see above). TLevel is the
// parent to all the TThings (Items, NPCs, and also the Player (who of
// course is also a NPC ;-) ).
type

{ TLevel }

TLevel = class(TGameObject)
       // The map array that holds TCells.
       Map  : array[1..MapSizeX,1..MapSizeY] of TCell;
       ShiftX : Integer;
       ShiftY : Integer;
       Level  : Byte;
       CellX,
       CellY  : Byte;
       // Overriden constructor.
       constructor Create(const thingID : string); override;
       // Disposes the children (TThings) and clears the map.
       procedure Clear;
       // Draws the level to screen.
       procedure Draw;
       // Clears the lighted and blocked bits from the lightmap.
       procedure LightClear;
       // Traces a light at (lx,ly) with the radius ldist.
       procedure LightTrace(lx,ly : Word; ldist : Byte);
       // Levels time flow procedure. See TGameObject.
       procedure TimeFlow(time : LongInt);
       // CalculatesShift value based on X,Y
       procedure CalcShift(px,py : Word);
       // Properly adds monster to a given square.
       procedure AddMonster(monster : string; x,y : byte);
       // Properly adds monster to a given square.
       procedure AddMonster(npc : TNPC; x,y : byte);
       // Properly adds item to a given square.
       procedure AddItem(const itemid : string; x,y : byte);
       // Properly adds item to a given square.
       procedure AddItem(item : TItem; x,y : byte);
       // Adds an item to the nearest free square.
       procedure DropItem(item : TItem; x,y : byte);
       // Adds an item to the nearest free square.
       procedure DropMonster(npc : TNPC; x,y : byte);
       // Adds an item to the nearest free square.
       procedure DropItem(const itemid : string; x,y : byte);
       // DiabloStyle -- draw level on the left part of screen
       procedure DrawLeft;
       // DiabloStyle -- draw level on the right part of screen
       procedure DrawRight;
       // Places a portal. Will check if portal exists and remove it.
       procedure PlacePortal(portalID,x,y : Byte);
       // Find given tile
       procedure FindCell(CellID : Byte);
       // Find nearest free square.
       procedure FindNearest(x,y : Byte; const cFlags : TFlags);
       // noMonsters,noItems,noObstacles
       function CellCheck(x,y : Byte; const cFlags : TFlags) : boolean;
       // checks wether given cell is visible
       function isVisible(x,y : Byte) : boolean;
       // Overriden destructor.
       destructor Destroy; override;
       private
       // Draws the cell at x,y.
       procedure Draw(x,y : Word);
     end;

const noMonsters  = 1;
      noItems     = 2;
      noObstacles = 3;
      noChangeRes = 4;

implementation
uses sysutils,vflags,voutput,vmath,rlmsg,rlcell,rlgen, rlgame;

constructor TLevel.Create(const thingID : string);
var ltype : string;
begin
  // With TNode descendants ALWAYS call inherited Init and Done.
  inherited Create(thingID);
  Log('Created.');
  ShiftX := 0;
  ShiftY := 0;
  ltype := Database.GetString('ltype');
  level := Database.GetNumber('level');
  Clear;
  GenerateLevel(Self,ltype);
  
  CellX := 0;
  CellY := 0;
end;

procedure TLevel.CalcShift(px,py : Word);
begin
  ShiftX := px-39;
  ShiftY := py-10;
  if ShiftX < 0 then ShiftX := 0;
  if ShiftY < 0 then ShiftY := 0;
  if 78+ShiftX > MapSizeX then ShiftX := MapSizeX-78;
  if 20+ShiftY > MapSizeY then ShiftY := MapSizeY-20;
end;

procedure TLevel.Draw(x,y : Word);
  function Color(Value : Byte) : Byte;
  begin
    if Value <= 16 then Exit(Value);
    case Value of
      ColorPortal : case PTurnCount mod 4 of 0 : Exit(Blue); 1,3 : Exit(LightBlue); 2 : Exit(White); end;
    else Exit(Red)
    end;
  end;
begin
  if not FlagSet(Map[x,y].Light,lfVisited) then Exit;
  if FlagSet(Map[x,y].Light,lfLighted) then
  begin
    Output.DrawChar(x+MapPosX-ShiftX,y+MapPosY-ShiftY,Color(Cells[Map[x,y].Cell].Color),Cells[Map[x,y].Cell].Pic);
    if Map[x,y].Item <> nil then Output.DrawChar(x+MapPosX-ShiftX,y+MapPosY-ShiftY,Map[x,y].Item.Color,Map[x,y].Item.Pic);
    if Map[x,y].NPC <> nil  then Output.DrawChar(x+MapPosX-ShiftX,y+MapPosY-ShiftY,Map[x,y].NPC.Color,Map[x,y].NPC.Pic);
  end
  else
    Output.DrawChar(x+MapPosX-ShiftX,y+MapPosY-ShiftY,DarkGray,Cells[Map[x,y].Cell].Pic);
end;

// Clears the lighted and blocked bits from the lightmap.
procedure Tlevel.LightClear;
var tx, ty : Integer;
begin
  for tx := 1 to MapSizeX do
    for ty := 1 to MapSizeY do
      MultiClearBFlag(Map[tx,ty].Light,[lfBlocked,lfLighted]);
end;

// Special thanks for this procedure goes to Isaac Kuo. This beamcasting
// algorithm is a ported to FreePascal modified version of his algorithm
// posted on http://www.roguelikedevelopment.org
//   lx,ly -- the coordinates of the lightsource
//   ldist -- the maximum range of the lightsource
procedure TLevel.LightTrace(lx,ly : Word; ldist : Byte);
var tx, ty : Integer;
    mini, maxi, cor, u, v : Integer;
    quad, slope : Byte;
const quads : array[1..4] of array[1..2] of ShortInt =
      ((1,1),(-1,-1),(-1,+1),(+1,-1));
      RayNumber          = 32; // It's effective to keep this number a power of 2.
      RayWidthCorrection = 10; // Must be smaller then RayNumber/2 ;-)
begin
  for tx := 1 to MapSizeX do
    for ty := 1 to MapSizeY do
      if (Cells[Map[tx,ty].Cell].bvis) or (Distance(lx,ly,tx,ty) > ldist) then
        SetBFlag(Map[tx,ty].Light,lfBlocked);
  // Set 0,0 to be visible even if the player is
  // standing on something opaque
  MultiSetBFlag(Map[lx,ly].Light,[lflighted,lfvisited]);

  // Check the orthogonal directions
  tx := lx; repeat Inc(tx); MultiSetBFlag(Map[tx,ly].Light,[lflighted,lfvisited]); until FlagSet(Map[tx,ly].Light,lfBlocked);
  tx := lx; repeat Dec(tx); MultiSetBFlag(Map[tx,ly].Light,[lflighted,lfvisited]); until FlagSet(Map[tx,ly].Light,lfBlocked);
  ty := ly; repeat Inc(ty); MultiSetBFlag(Map[lx,ty].Light,[lflighted,lfvisited]); until FlagSet(Map[lx,ty].Light,lfBlocked);
  ty := ly; repeat Dec(ty); MultiSetBFlag(Map[lx,ty].Light,[lflighted,lfvisited]); until FlagSet(Map[lx,ty].Light,lfBlocked);

  // Loop through the quadrants
  for quad := 1 to 4 do
  // Now loop on the diagonal directions
  for slope := 1 to RayNumber-1 do
  begin
    // initialize the v coordinate and set the beam size
    // to maximum--mini and maxi store the beam\'s current
    // top and bottom positions.
    // As long as mini<maxi, the beam has some width.
    // When mini=maxi, the beam is a thin line.
    // When mini>maxi, the beam has been blocked.

    v := slope; u := 0;
    mini := RayWidthCorrection; maxi := RayNumber-RayWidthCorrection;
    repeat Inc(u);
      ty:= v div RayNumber; tx:= u - ty;  //Do the transform
      cor:= RayNumber-(v mod RayNumber);         //calculate the position of block corner within beam
      if mini < cor then begin //beam is low enough to hit (x,y) block
        MultiSetBFlag(Map[quads[quad][1]*tx+lx,quads[quad][2]*ty+ly].Light,[lflighted,lfvisited]);
        if FlagSet(Map[quads[quad][1]*tx+lx,quads[quad][2]*ty+ly].Light,lfBlocked) then mini := cor; //beam was partially blocked
      end;
      if maxi > cor then begin //beam is high enough to hit (x-1,y+1) block
        MultiSetBFlag(Map[lx+quads[quad][1]*(tx-1),ly+quads[quad][2]*(ty+1)].Light,[lflighted,lfvisited]);
        if FlagSet(Map[lx+quads[quad][1]*(tx-1),ly+quads[quad][2]*(ty+1)].Light,lfBlocked) then maxi := cor; //beam was partially blocked
      end;
      v := v + slope;  //increment the beam\'s v coordinate
    until mini > maxi;
  end;
end;

procedure TLevel.DrawRight;
var dx,dy,mx : Word;
    sx       : Integer;
begin
  sx := ShiftX; mx := MapPosX;
  ShiftX  := Game.Player.X-20;
  MapPosX := 40;
  if ShiftX < 0 then ShiftX := 0;
  if 39+ShiftX > MapSizeX then ShiftX := MapSizeX-39;
  for dx := ShiftX+1 to Min(ShiftX+39,MapSizeX) do
    for dy := ShiftY+1 to Min(ShiftY+20,MapSizeY) do
      Draw(dx,dy);
  ShiftX := sx; MapPosX := mx;
end;

procedure TLevel.PlacePortal(portalID, x, y : Byte);
begin
  FindCell(portalID);
  if CellX <> 0 then Map[CellX,CellY].Cell := CellFloor;
  FindNearest(x,y,[noMonsters,noObstacles,noItems,noChangeRes]);
  if CellX = 0 then Exit;
  Map[CellX,CellY].Cell := portalID;
  if isVisible(CellX,CellY) then Msg.Add('A portal appears!');
end;

procedure TLevel.FindCell(CellID : Byte);
var x,y : Byte;
begin
  for x := 1 to MapSizeX do
    for y := 1 to MapSizeY do
      if Map[x,y].Cell = CellID then
      begin
        CellX := x;
        CellY := y;
        Exit;
      end;
  CellX := 0;
  CellY := 0;
end;

procedure TLevel.FindNearest(x, y : Byte; const cFlags : TFlags);
var cn,cx,cy : byte;
begin
  if CellCheck(x,y,cFlags) then begin CellX := x; CellY := y; exit; end;
  for cn := 1 to 8 do
  begin
    CellX := x+Random(3)-1;
    CellY := y+Random(3)-1;
    if CellCheck(CellX,CellY,cFlags) then exit;
  end;
  for cn := 2 to 40 do
  begin
    for cx := Max(x-cn,1) to Min(x+cn,MapSizeX) do
      for cy := Max(y-cn,1) to Min(y+cn,MapSizeY) do
        if CellCheck(cx,cy,cFlags) then
        begin
          CellX := cx;
          CellY := cy;
          exit;
        end;
  end;
  // if we couldn't place the item we need to destroy it :-(
  CellX := 0;
  CellY := 0;
end;

function TLevel.CellCheck(x, y : Byte; const cFlags : TFlags) : boolean;
begin
  if (noMonsters  in cFlags) and (Map[x,y].NPC         <> nil)  then Exit(False);
  if (noItems     in cFlags) and (Map[x,y].Item        <> nil)  then Exit(False);
  if (noObstacles in cFlags) and (not Cells[Map[x,y].Cell].Pass)  then Exit(False);
  if (noChangeRes in cFlags) and (not Cells[Map[x,y].Cell].canch) then Exit(False);
  Exit(True);
end;

function TLevel.isVisible(x, y : Byte) : boolean;
begin
  Exit(FlagSet(Map[x,y].Light,lfLighted))
end;

procedure TLevel.DrawLeft;
var dx,dy : Word;
    sx    : Integer;
begin
  sx := ShiftX;
  ShiftX  := Game.Player.X-20;
  if ShiftX < 0 then ShiftX := 0;
  if 39+ShiftX > MapSizeX then ShiftX := MapSizeX-39;
  for dx := ShiftX+1 to Min(ShiftX+39,MapSizeX) do
    for dy := ShiftY+1 to Min(ShiftY+20,MapSizeY) do
      Draw(dx,dy);
  ShiftX := sx;
end;


procedure TLevel.Draw;
var dx,dy : Word;
begin
  for dx := ShiftX+1 to Min(ShiftX+78,MapSizeX) do
    for dy := ShiftY+1 to Min(ShiftY+20,MapSizeY) do
      Draw(dx,dy);
end;

procedure TLevel.TimeFlow(time : LongInt);
begin
  // Do things.
  with Game.Player do
    if PortalCount > 0 then
       if PortalCount = 1 then
       begin
         PortalLevel := LevelNumber;
         PlacePortal(CellTownPortal,x,y);
         PortalCount := 0;
       end else Dec(PortalCount);
  // Pass time flow to children:
  inherited TimeFlow(time);
end;

procedure TLevel.Clear;
var cx, cy : byte;
begin
  // First destroy all children.
  if Child <> nil then FreeAndNil(Child);

  // Now clear the map.
  for cx := 1 to MapSizeX do
    for cy := 1 to MapSizeY do
      with Map[cx,cy] do
      begin
        Cell   := CellFloor;
        Light  := 0;
        NPC    := nil;
        Item   := nil;
      end;
end;

// Properly adds monster to a given square.
procedure TLevel.AddMonster(monster : string; x,y : byte);
var NPC : TNPC;
begin
  NPC := TNPC.Create(monster);
  Add(NPC);
  NPC.Displace(x,y);
end;

// Properly adds item to a given square.
procedure TLevel.AddItem(const itemid : string; x,y : byte);
begin
  AddItem(TItem.Create(itemid),x,y);
end;

// Properly adds item to a given square.
procedure TLevel.AddItem(item : TItem; x,y : byte);
begin
  Add(Item);
  Item.Displace(x,y);
end;

// Properly adds item to a given square.
procedure TLevel.DropItem(const itemid : string; x,y : byte);
begin
  DropItem(TItem.Create(itemid),x,y);
end;

// Adds an item to the nearest free square.
procedure TLevel.DropItem(item : TItem; x,y : byte);
begin
  FindNearest(x,y,[noItems,noObstacles]);
  if CellX = 0 then begin FreeAndNil(Item); exit end;
  AddItem(item,CellX,CellY);
end;

// Properly adds monster to a given square.
procedure TLevel.AddMonster(npc : TNPC; x,y : byte);
begin
  Add(NPC);
  NPC.Displace(x,y);
end;

// Adds an item to the nearest free square.
procedure TLevel.DropMonster(npc : TNPC; x,y : byte);
begin
  FindNearest(x,y,[noMonsters,noObstacles]);
  if CellX = 0 then begin FreeAndNil(npc); exit end;
  AddMonster(npc,CellX,CellY);
end;


destructor TLevel.Destroy;
begin
  // With TNode descendants ALWAYS call inherited Init and Done.
  inherited Destroy;
end;

end.
