unit vlua;
interface
uses vnode,vutil,lua,classes,vconsole,vdf, lualib, lauxlib;

type
  ELuaException       = class(EException);

{ TLua }

TLua = class(TVObject)
               constructor Create; virtual;
               procedure LoadFile(const FileName : string);
               procedure StreamLoader(IST : TStream; StreamName : AnsiString;  Size : DWord);
               procedure LoadStream(DF : TVDataFile; const StreamName : string);
               procedure AddVar(const Name : string; const l : LongInt);
               procedure AddVar(const Name : string; const s : string);
               procedure GetVar(const Name : string; out l : LongInt);
               procedure GetVar(const Name : string; out s : string);
               function FunctionExists(const Name : string) : boolean;
               function Execute(const Name : string) : boolean;
               function Execute(const Name : string; const Number : LongInt) : boolean;
//               function Execute(const Name : string; Args : Array of Const) : boolean;
               function ExecuteCommand(const Command : string) : boolean;
               function ExecuteCommand(const Command : string; const Number : LongInt) : boolean;
               procedure Register(const Name : string; Proc : lua_CFunction);
               procedure RegisterConst(const Name : string; value : Integer);
               procedure Error(const ErrorString : Ansistring); virtual;
               destructor Destroy; override;
               LuaState : Plua_State;
               ErrorStr : AnsiString;
               protected
               Ansi     : AnsiString;
             end;
TLuaTable = class(TVObject)
  constructor Create( aLua : TLua; aName : AnsiString );
  constructor Create( aLua : TLua; aName : AnsiString; aIndex : AnsiString );
  constructor Create( aLua : TLua; aName : AnsiString; aIndex : LongInt );
  function getNumber(const ID : AnsiString) : LongInt;
  function getFloat(const ID : AnsiString) : Real;
  function getString(const ID : AnsiString) : string;
  function getBoolean(const ID : AnsiString) : boolean;
  function isFunction(const ID : AnsiString) : boolean;
  procedure Execute(const ID : AnsiString; Params : array of Const);
  procedure Execute(const ID : AnsiString);
  destructor Destroy; override;
private
  mLua  : TLua;
  mName : AnsiString;
  mPop  : Byte;
end;


type

{ TLuaConsole }

TLuaConsole = class(TConsole)
  constructor Create(ConsoleWindow : TConsoleWindow; nLua : TLua);
  procedure Command(cmd : Ansistring); override;
  private
  VLua : TLua;
end;

implementation
uses SysUtils;

constructor TLua.Create;
begin
  LuaState := lua_open;
  luaopen_base(LuaState);
end;

procedure TLua.LoadFile(const FileName : string);
begin
  Ansi := Filename;
  if lua_dofile(LuaState, PChar(Ansi)) <> 0 then
    CritError(lua_tostring(LuaState,-1));
end;

procedure TLua.AddVar(const Name : string; const l : LongInt);
begin
  Ansi := Name;
  lua_pushnumber(LuaState,l);
  lua_setglobal(LuaState,PChar(Ansi));
end;

procedure TLua.GetVar(const Name : string; out l : LongInt);
begin
  Ansi := Name;
  lua_getglobal(LuaState,PChar(Ansi));
  l := Round(lua_tonumber(LuaState,-1));
  lua_pop(LuaState,1);
end;

procedure TLua.GetVar(const Name : string; out s : string);
begin
  Ansi := Name;
  lua_getglobal(LuaState,PChar(Ansi));
  Ansi := lua_tostring(LuaState,-1);
  s := ansi;
  lua_pop(LuaState,1);
end;

procedure TLua.AddVar(const Name : string; const s : String);
begin
  Ansi := s;
  lua_pushstring(LuaState,PChar(Ansi));
  Ansi := Name;
  lua_setglobal(LuaState,PChar(Ansi));
end;

procedure TLua.LoadStream(DF: TVDataFile; const StreamName: string);
var Stream : TStream;
    Size   : Int64;
begin
  Ansi := StreamName;
  Stream := DF.GetFile(Ansi);
  Size   := DF.GetFileSize(Ansi);
  StreamLoader(Stream,Ansi,Size);
  FreeAndNil(Stream);
end;

procedure TLua.StreamLoader(IST : TStream; StreamName : AnsiString;  Size : DWord);
var Buf  : PByte;
begin
  Log('Loading LUA stream -- "'+StreamName+'" ('+IntToStr(Size)+'b)');
  GetMem(Buf,Size);
  Log('Reading "'+StreamName+'" ('+IntToStr(IST.Position)+'-'+IntToStr(IST.Position+Size)+')');
  IST.ReadBuffer(Buf^,Size);
  if luaL_loadbuffer(LuaState,PChar(Buf),Size,PChar(StreamName)) = 0 then
    if lua_pcall(LuaState, 0, 0, 0)  <> 0 then
    begin
      ErrorStr := lua_tostring(LuaState,-1);
      Error(StreamName+': '+ErrorStr);
      lua_pop(LuaState,1);
    end;
  FreeMem(Buf);
  Log('Loaded "'+StreamName+'" ('+IntToStr(Size)+'b)');
end;


function TLua.FunctionExists(const Name : string) : boolean;
begin
  Ansi := Name;
  lua_getglobal(LuaState, PChar(Ansi));
  FunctionExists := lua_type(LuaState,-1) = LUA_TFUNCTION;
  lua_pop(LuaState,1);
end;

function TLua.Execute(const Name : string) : boolean;
begin
  Ansi := Name;
  lua_getglobal(LuaState, PChar(Ansi));
  if lua_pcall(LuaState, 0,0,0) <> 0 then
  begin
    ErrorStr := lua_tostring(LuaState,-1);
    Error(Name+': '+ErrorStr);
    lua_pop(LuaState,1);
    Exit(False);
  end;
  Exit(True);
end;

function TLua.Execute(const Name : string; const Number : LongInt) : boolean;
begin
  Ansi := Name;
  lua_getglobal(LuaState, PChar(Ansi));
  lua_pushnumber(LuaState,Number);
  if lua_pcall(LuaState, 1,0,0) <> 0 then
  begin
    ErrorStr := lua_tostring(LuaState,-1);
    Error(Name+': '+ErrorStr);
    lua_pop(LuaState,1);
    Exit(False);
  end;
  Exit(True);
end;

function TLua.ExecuteCommand(const Command : string; const Number : LongInt) : boolean;
var Args,Arg : string;
    Count    : Byte;
begin
  Ansi := Parameter(Command,1,'(');
  Args := Parameter(Command,1,'()');

  if Args = '' then Exit(Execute(Ansi,Number));

  lua_getglobal(LuaState, PChar(Ansi));
  lua_pushnumber(LuaState,Number);

  Count := 1;
  repeat
    ///////////////////// MAY BE OPTIMIZED USING SPLIT INSTEAD OF Parameter
    Arg := Parameter(Args,Count,',');

    ///////////////////// NO  ERROR CHECKING!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    if Arg = '' then Break;
    Arg := Trim(Arg);
    case Arg[1] of
      '-','0'..'9' : lua_pushnumber(LuaState,StrToInt(Arg));
      'T','t'      : lua_pushboolean(LuaState,true);
      'F','f'      : lua_pushboolean(LuaState,false);
      ''''         : begin Ansi := Parameter(Arg,1,''''''); lua_pushstring(LuaState,PChar(Ansi)); end;
      '"'          : begin Ansi := Parameter(Arg,1,'""'); lua_pushstring(LuaState,PChar(Ansi)); end;
      else Log('Invalid command arguments passed to ExecuteCommand : '+Command);
    end;
    Inc(Count);
  until Arg = '';

  if lua_pcall(LuaState, Count,0,0) <> 0 then
  begin
    ErrorStr := lua_tostring(LuaState,-1);
    Error(Command+': '+ErrorStr);
    lua_pop(LuaState,1);
    Exit(False);
  end;
  Exit(True);
end;

function TLua.ExecuteCommand(const Command : string) : boolean;
var Args,Arg : string;
    Count    : Byte;
begin
  Ansi := Parameter(Command,1,'(');
  Args := Parameter(Command,1,'()');

  if Args = '' then Exit(Execute(Ansi));

  lua_getglobal(LuaState, PChar(Ansi));

  Count := 1;
  repeat
    ///////////////////// MAY BE OPTIMIZED USING SPLIT INSTEAD OF Parameter
    Arg := Parameter(Args,Count,',');

    ///////////////////// NO  ERROR CHECKING!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    if Arg = '' then Break;
    Arg := Trim(Arg);
    case Arg[1] of
      '-','0'..'9' : lua_pushnumber(LuaState,StrToInt(Arg));
      'T','t'      : lua_pushboolean(LuaState,true);
      'F','f'      : lua_pushboolean(LuaState,false);
      ''''         : begin Ansi := Parameter(Arg,1,''''''); lua_pushstring(LuaState,PChar(Ansi)); end;
      '"'          : begin Ansi := Parameter(Arg,1,'""'); lua_pushstring(LuaState,PChar(Ansi)); end;
      else Log('Invalid command arguments passed to ExecuteCommand : '+Command);
    end;
    Inc(Count);
  until Arg = '';

  if lua_pcall(LuaState, Count-1,0,0) <> 0 then
  begin
    ErrorStr := lua_tostring(LuaState,-1);
    Error(Command+': '+ErrorStr);
    lua_pop(LuaState,1);
    Exit(False);
  end;
  Exit(True);
end;


{function Execute(const Name : string; Args : Array of Const) : boolean;
var count : Byte;
begin
  if High(Args) < 0 then Exit(Execute(Name));
  Ansi := Name;
  lua_getglobal(LuaState, PChar(Ansi));

  for count := 0 to High(Args) do
  case
  lua_pushnumber(LuaState,Number);



  if lua_pcall(LuaState, High(Args)+1,0,0) <> 0 then
  begin
    Error := lua_tostring(LuaState,-1);
    Log('ERROR/'+Name+': '+Error);
    lua_pop(LuaState,1);
    Exit(False);
  end;
end;  }

procedure TLua.Register(const Name : string; Proc : lua_CFunction);
begin
  Ansi := Name;
  lua_register(LuaState, PChar(Ansi), Proc);
end;

procedure TLua.RegisterConst(const Name : string; value : Integer);
begin
  Ansi := Name;
  lua_pushstring(LuaState,PChar(Ansi));
  lua_pushnumber(LuaState,value);
  lua_rawset(LuaState,LUA_GLOBALSINDEX);
end;

procedure TLua.Error(const ErrorString: Ansistring);
begin
  Log('LuaError: '+ErrorString);
end;

destructor TLua.Destroy;
begin
  lua_close(LuaState);
  Log('Lua closed.');
  inherited Destroy;
end;

function lua_print(L: Plua_State): Integer; cdecl;
var n : Integer;
begin
  n := lua_gettop(L);
  if n < 0 then Exit(0);
  Console.Print(lua_tostring(L,1));
  Result := 0;
end;

constructor TLuaConsole.Create(ConsoleWindow: TConsoleWindow; nLua: TLua);
begin
  inherited Create(ConsoleWindow);
  VLua := nLua;
  VLua.Register('print',@lua_print);
end;

procedure TLuaConsole.Command(cmd : Ansistring);
var Error : AnsiString;
begin
  cmd := Trim(cmd);
  Print('@B> @l'+cmd);
  if lua_dostring(VLua.LuaState,PChar(cmd)) <> 0 then
  begin
    Error := lua_tostring(VLua.LuaState,-1);
    Print('@RError: @l'+Error);
    lua_pop(VLua.LuaState,1);
  end;
end;

{ TLuaTable }

constructor TLuaTable.Create(aLua: TLua; aName: AnsiString);
begin
  mName := aName;
  mLua  := aLua;
  lua_getglobal( mLua.LuaState, PChar( mName ) );
  if not lua_istable( mLua.LuaState, -1) then ELuaException.Create(mName+' is not a valid table!');
  mPop := 1;
end;

constructor TLuaTable.Create(aLua: TLua; aName: AnsiString; aIndex : AnsiString);
begin
  Create( aLua, aName );
  lua_pushstring( mLua.LuaState, PChar( aIndex ) );
  lua_gettable( mLua.LuaState, -2 );  // get tid[key]
  if not lua_istable(mLua.LuaState, -1) then ELuaException.Create('@1[@2] not found!', [mName,aIndex]);
  mPop := 2;
end;

constructor TLuaTable.Create(aLua: TLua; aName: AnsiString; aIndex: LongInt);
begin
  Create( aLua, aName );
  lua_pushnumber( mLua.LuaState, aIndex );
  lua_gettable( mLua.LuaState, -2 );  // get tid[key]
  if not lua_istable(mLua.LuaState, -1) then ELuaException.Create('@1[@2] not found!', [mName,aIndex]);
  mPop := 2;
end;

function TLuaTable.getNumber(const ID: AnsiString): LongInt;
begin
  lua_pushstring( mLua.LuaState, PChar(ID));
  lua_gettable(mLua.LuaState, -2);  // get background[key] */
  if (not lua_isnumber(mLua.LuaState, -1))
    then getNumber := 0
    else getNumber := Round(lua_tonumber(mLua.LuaState, -1));
  lua_pop(mLua.LuaState, 1);
end;

function TLuaTable.getFloat(const ID: AnsiString): Real;
begin
  lua_pushstring(mLua.LuaState, PChar(ID));
  lua_gettable(mLua.LuaState, -2);  // get background[key] */
  if (not lua_isnumber(mLua.LuaState, -1))
    then getFloat := 0
    else getFloat := lua_tonumber(mLua.LuaState, -1);
  lua_pop(mLua.LuaState, 1);
end;

function TLuaTable.getString(const ID: AnsiString): string;
begin
  lua_pushstring(mLua.LuaState, PChar(ID));
  lua_gettable(mLua.LuaState, -2);  // get background[key]
  if lua_isstring(mLua.LuaState,-1) or lua_isnumber(mLua.LuaState,-1) then
    getString := lua_tostring(mLua.LuaState,-1)
  else
    getString := '';
  lua_pop(mLua.LuaState, 1);
end;

function TLuaTable.getBoolean(const ID: AnsiString): boolean;
begin
  lua_pushstring(mLua.LuaState, PChar(ID));
  lua_gettable(mLua.LuaState, -2);  // get background[key] */
  getBoolean := lua_toboolean(mLua.LuaState, -1);
  lua_pop(mLua.LuaState, 1);
end;

function TLuaTable.isFunction(const ID: AnsiString): boolean;
begin
  lua_pushstring(mLua.LuaState, PChar(ID));
  lua_gettable(mLua.LuaState, -2);  // get table[ID] */
  isFunction := lua_isfunction(mLua.LuaState, -1);
  lua_pop(mLua.LuaState, 1);
end;

procedure TLuaTable.Execute(const ID: AnsiString; Params: array of const);
var Args, Count : Integer;
    Temp : Ansistring;
begin
  Args := High(Params)+1;
  lua_pushstring(mLua.LuaState, PChar(ID));
  lua_gettable(mLua.LuaState, -2);  // get Table[ID] */
  if not lua_isfunction(mLua.LuaState, -1) then ELuaException.Create(ID+' not found in table '+mName+'!');
  if Args > 0 then
  for Count := 1 to Args do
  begin
    case Params[Count-1].vtype of
      vtInteger    : lua_pushnumber(mLua.LuaState,params[Count-1].vinteger);
      vtBoolean    : lua_pushboolean(mLua.LuaState,params[Count-1].vboolean);
      vtChar       : begin Temp := params[Count-1].vchar; lua_pushstring(mLua.LuaState,PChar(Temp)); end;
      vtString     : lua_pushstring(mLua.LuaState,PChar(params[Count-1].vstring));
      vtPChar      : lua_pushstring(mLua.LuaState,PChar(params[Count-1].vpchar));
      vtAnsiString : lua_pushstring(mLua.LuaState,PChar(params[Count-1].vansistring));
      vtExtended   : lua_pushnumber(mLua.LuaState,params[Count-1].vextended^)
      else ELuaException.Create('Unknown type passed to TableExecute!');
    end;
  end;
  if lua_pcall(mLua.LuaState, Args,0,0) <> 0 then
  begin
    mLua.ErrorStr := lua_tostring(mLua.LuaState,-1);
    mLua.Error(mName+'.'+ID+': '+mLua.ErrorStr);
    lua_pop(mLua.LuaState,1);
  end;
end;

procedure TLuaTable.Execute(const ID: AnsiString);
begin
  Execute(ID,[]);
end;



destructor TLuaTable.Destroy;
begin
  lua_pop( mLua.LuaState, mPop );
  inherited Destroy;
end;

end.

